import codecs
import functools
from eth_utils import is_hex
from rlp.utils import (
    big_endian_to_int,
    int_to_big_endian,
    encode_hex,
    decode_hex as _decode_hex,
)


def bytearray_to_bytestr(value):
    return bytes(value)


def is_numeric(value):
    return isinstance(value, int)


def is_binary(value):
    return isinstance(value, (bytes, bytearray))


def is_text(value):
    return isinstance(value, str)


def is_string(value):
    return isinstance(value, (bytes, str, bytearray))


def is_integer(value):
    return isinstance(value, int)


def is_array(value):
    return isinstance(value, (list, tuple))


def force_text(value):
    if is_text(value):
        return value
    elif is_binary(value):
        return codecs.decode(value, "iso-8859-1")
    else:
        raise TypeError("Unsupported type: {0}".format(type(value)))


def force_bytes(value):
    if is_binary(value):
        return bytes(value)
    elif is_text(value):
        return codecs.encode(value, "iso-8859-1")
    else:
        raise TypeError("Unsupported type: {0}".format(type(value)))


def force_obj_to_text(obj, skip_unsupported=False):
    if is_string(obj):
        return force_text(obj)
    elif isinstance(obj, dict):
        return {
            k: force_obj_to_text(v, skip_unsupported) for k, v in obj.items()
        }
    elif isinstance(obj, (list, tuple)):
        return type(obj)(force_obj_to_text(v, skip_unsupported) for v in obj)
    elif not skip_unsupported:
        raise ValueError("Unsupported type: {0}".format(type(obj)))
    else:
        return obj


def force_obj_to_bytes(obj, skip_unsupported=False):
    if is_string(obj):
        return force_bytes(obj)
    elif isinstance(obj, dict):
        return {
            k: force_obj_to_bytes(v, skip_unsupported) for k, v in obj.items()
        }
    elif isinstance(obj, (list, tuple)):
        return type(obj)(force_obj_to_bytes(v, skip_unsupported) for v in obj)
    elif not skip_unsupported:
        raise ValueError("Unsupported type: {0}".format(type(obj)))
    else:
        return obj


def coerce_args_to_bytes(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        bytes_args = force_obj_to_bytes(args, True)
        bytes_kwargs = force_obj_to_bytes(kwargs, True)
        return fn(*bytes_args, **bytes_kwargs)
    return inner


def coerce_return_to_bytes(fn):
    @functools.wraps(fn)
    def inner(*args, **kwargs):
        return force_obj_to_bytes(fn(*args, **kwargs), True)
    return inner


@coerce_args_to_bytes
def strip_0x(value):
    if value.startswith(b'0x'):
        return value[2:]
    return value


@coerce_args_to_bytes
def add_0x(value):
    return b"0x" + strip_0x(value)


@coerce_return_to_bytes
def encode_data(data, length=None):
    '''Encode unformatted binary `data`.

    If `length` is given, the result will be padded like this: ``quantity_encoder(255, 3) ==
    '0x0000ff'``.
    '''

    def zpad(x, l):
        ''' Left zero pad value `x` at least to length `l`.

        >>> zpad('', 1)
        '\x00'
        >>> zpad('\xca\xfe', 4)
        '\x00\x00\xca\xfe'
        >>> zpad('\xff', 1)
        '\xff'
        >>> zpad('\xca\xfe', 2)
        '\xca\xfe'
        '''
        return b'\x00' * max(0, l - len(x)) + x
    return add_0x(encode_hex(zpad(data, length or 0)))


@coerce_return_to_bytes
def encode_number(value, length=None):
    '''Encode interger quantity `data`.'''
    if not is_numeric(value):
        raise ValueError("Unsupported type: {0}".format(type(value)))
    hex_value = encode_data(int_to_big_endian(value), length)

    if length:
        return hex_value
    else:
        return add_0x(strip_0x(hex_value).lstrip(b'0') or b'0')


@coerce_args_to_bytes
def decode_hex(value):
    return _decode_hex(strip_0x(value))


@coerce_args_to_bytes
def normalize_number(value):
    if is_numeric(value):
        return value
    elif is_string(value):
        if value.startswith(b'0x'):
            return int(value, 16)
        else:
            return big_endian_to_int(value)
    else:
        raise ValueError("Unknown numeric encoding: {0}".format(value))


def normalize_block_identifier(block_identifier):
    if type(block_identifier) == dict:
        for k, v in block_identifier.items():
            if k == "blockHash":
                print("blockHash value", v)
                return v
            if k == "blockNumber":
                print("blockNumber value", v)
                return v
        raise ValueError("Unknown block identifier encoding: {0}".format(block_identifier))

    if is_hex(block_identifier):
        return block_identifier
    if block_identifier is None or block_identifier == "earliest":
        return normalize_number(0)
    if block_identifier in ["best", "latest", "pending"]:  # eth 最新块用 latest 表示
        return "best"
    return normalize_number(block_identifier)



import aiohttp
import json

async def post(endpoint_uri, data, **kwargs):
    async with aiohttp.ClientSession() as session:
        async with session.post(endpoint_uri, json=data, **kwargs) as response:
            print("Raw response from server: ", response.status, await response.text())
            ctype = response.headers['Content-Type']
            expect = "text/plain"
            if expect in ctype:
                await response.text()
                return response
            else:
                await response.json()
                return response

async def get(endpoint_uri, params, **kwargs):
    async with aiohttp.ClientSession() as session:
        async with session.get(endpoint_uri, params=params, **kwargs) as response:
            print("Raw response from server: ", response.status,  await response.text())
            ctype = response.headers['Content-Type']
            expect = "text/plain"
            if expect in ctype:
                await response.text()
                return response
            else:
                await response.json()
                return response

class Restful(object):

    def __init__(self, endpoint):
        super(Restful, self).__init__()
        self._endpoint = endpoint

    def __call__(self, parameter):
        if parameter is not None:
            return Restful('%s/%s' % (self._endpoint, parameter))
        return self

    def __getattr__(self, resource):
        return Restful('%s/%s' % (self._endpoint, resource))

    async def make_request(self, method, params=None, data=None, **kwargs):
        headers = {
            "accept": "application/json",
            "Connection": "keep-alive",
            "Content-Type": "application/json"
        }
        kwargs.setdefault('headers', headers)
        kwargs.setdefault('timeout', 10)
        kwargs.setdefault('chunked', True)
        error = None
        try:
            response = await method(self._endpoint, params=params, data=data, **kwargs)
            return await response.json()
        except aiohttp.ClientConnectionError as e:
            print("Unable to connect to Meter-Restful server:")
            error = e
            raise e
        except Exception as e:
            text = ''
            try:
                text = await response.text()
                error = Exception(text.strip('\n'))
            except:
                error = e
                raise e
        print("Meter-Restful server Err:")
        print("Error message from meter node:", error)
        #raise error
        x ='{"id": "' + text.strip('\n') + '"}' 
        #print(x)
        return json.loads(x)



import itertools
import functools
import logging
import sys
import json
import traceback
import random
import time
import websockets
import asyncio
import hashlib
import copy

from .meter.client import meter
from .utils.compat import noop
from .utils.types import (
    encode_number,
    force_obj_to_text,
    normalize_block_identifier,
    normalize_number
)
from jsonrpcserver import method




def async_serialize(func):
    @functools.wraps(func)
    async def wrapper(*args, **kw):
        try:
            result = await func(*args, **kw)
            if isinstance(result, str):
                return result
            return force_obj_to_text(result, True)
        except Exception as e:
            traceback.print_exc()
            raise e
    return wrapper


#
# formatter
#
TXN_FORMATTERS = {
    'value': normalize_number,
    'gas': normalize_number,
}


def input_transaction_formatter(transaction):
    return {
        k: TXN_FORMATTERS.get(k, noop)(v)
        for k, v in transaction.items()
    }


def input_log_filter_formatter(filter_params):
    params_range = {"unit": "block"}
    params_range["from"] = int(filter_params["fromBlock"], 16)
    to_blk = filter_params.get("toBlock", None)
    if to_blk:
        params_range["to"] = int(to_blk, 16)
    return {
        "range": params_range,
        "topicSets": topics_formatter(filter_params.get("topics", []))
    }


def topics_formatter(eth_topics):
    if eth_topics:
        matrix = [x if isinstance(x, list) else [x] for x in eth_topics]
        return [
            {
                "topic{}".format(index): topic
                for index, topic in enumerate(e)
            }
            for e in itertools.product(*matrix)
        ]
    return []


#
#
#
@method
async def rpc_modules():
    return {
        "eth": "1.0",
        "net": "1.0",
        "web3": "1.0",
    }


#
# debug
#
@method
@async_serialize
async def debug_traceTransaction(tx_hash, params):
    return await meter.trace_transaction(tx_hash)


@method
@async_serialize
async def debug_storageRangeAt(blk_hash, tx_index, contract_addr, key_start, max_result):
    return await meter.storage_range_at(blk_hash, tx_index, contract_addr, key_start, max_result)


#
# net_api
#
@method
async def net_version():
    return str(int(meter.get_chainid(), 16))


@method
async def net_listening():
    return False

#
# evm_api
# 没有真实实现, 只是为了实现接口
#


@method
async def evm_snapshot():
    return encode_number(0)


@method
async def evm_revert(snapshot_idx=None):
    return True


#
# web3
#
def make_version():
    from . import __version__
    return "Meter-Gear/" + __version__ + "/{platform}/python{v.major}.{v.minor}.{v.micro}".format(
        v=sys.version_info,
        platform=sys.platform,
    )


@method
async def web3_clientVersion():
    return make_version()


#
# eth_api
#
@method
@async_serialize
async def eth_getStorageAt(address, position, block_identifier="best"):
    if position.startswith("0x"):
        position = position[2:]
    position = "0x{}".format(position.zfill(64))
    return await meter.get_storage_at(
        address, position, normalize_block_identifier(block_identifier))


@method
@async_serialize
async def eth_chainId():
    return meter.get_chainid()


@method
@async_serialize
async def eth_gasPrice():
    return hex(int(500e9))  # default gas price set to 500 GWei


@method
@async_serialize
async def eth_getTransactionCount(address, block_identifier="best"):
    '''
    ethereum 用来处理 nonce, Meter 不需要
    '''
    # return encode_number(0)
    random.seed(time.time())
    nonce = random.randint(1, 0xffffffff)
    return encode_number(nonce)


@method
async def eth_accounts():
    accounts = meter.get_accounts()
    return accounts


@method
@async_serialize
async def eth_getCode(address, block_identifier="best"):
    return await meter.get_code(address, normalize_block_identifier(block_identifier))


@method
@async_serialize
async def eth_blockNumber():
    return encode_number(await meter.get_block_number())


@method
@async_serialize
async def eth_estimateGas(transaction):
    formatted_transaction = input_transaction_formatter(transaction)
    print("FORMATED TX:", formatted_transaction)
    result = await meter.estimate_gas(formatted_transaction)
    return encode_number(result)




@method
@async_serialize
async def eth_call(transaction, block_identifier="best"):
    formatted_transaction = input_transaction_formatter(transaction)
    return await meter.call(formatted_transaction, normalize_block_identifier(block_identifier))


@method
@async_serialize
async def eth_sendTransaction(transaction):
    '''
    发送未签名的交易
    '''
    formatted_transaction = input_transaction_formatter(transaction)
    return await meter.send_transaction(formatted_transaction)


@method
@async_serialize
async def eth_sign(transaction):
    '''
    发送未签名的交易
    '''
    return await {}


@method
@async_serialize
async def eth_sendRawTransaction(raw):
    '''
    发送已签名的交易
    '''
    return await meter.send_raw_transaction(raw)


@method
@async_serialize
async def eth_getBalance(address, block_identifier="best"):
    return await meter.get_balance(address, normalize_block_identifier(block_identifier))


@method
@async_serialize
async def eth_getTransactionByHash(tx_hash):
    if tx_hash:
        return await meter.get_transaction_by_hash(tx_hash)
    return None


@method
@async_serialize
async def eth_getTransactionReceipt(tx_hash):
    if tx_hash:
        return await meter.get_transaction_receipt(tx_hash)
    return None


@method
@async_serialize
async def eth_getBlockByHash(block_hash, full_tx=False):
    return await get_block(block_hash, full_tx)


@method
@async_serialize
async def eth_getBlockByNumber(block_number, full_tx=False):
    return await get_block(block_number, full_tx)


async def get_block(block_identifier, full_tx):
    blk = await meter.get_block(normalize_block_identifier(block_identifier))
    if blk and full_tx:
        blk["transactions"] = [await eth_getTransactionByHash(
            tx) for tx in blk["transactions"]]
    return blk


@method
async def eth_newBlockFilter():
    return await meter.new_block_filter()


@method
@async_serialize
async def eth_uninstallFilter(filter_id):
    return meter.uninstall_filter(filter_id)


@method
@async_serialize
async def eth_getFilterChanges(filter_id):
    return await meter.get_filter_changes(filter_id)


@method
@async_serialize
async def eth_getLogs(filter_obj):
    to_blk = filter_obj.get('toBlock', None)
    if (to_blk == 'latest'):
        latest = await meter.get_block('best')
        filter_obj['toBlock'] = latest['number']
    print(filter_obj)
    return await meter.get_logs(filter_obj.get("address", None), input_log_filter_formatter(filter_obj))

@method
@async_serialize
async def trace_filter(filter_obj):
    to_blk = filter_obj.get('toBlock', None)
    if (to_blk == 'latest'):
        latest = await meter.get_block('best')
        filter_obj['toBlock'] = latest['number']
    return await meter.get_trace_filter(filter_obj)

@method
@async_serialize
async def trace_transaction(filter_obj):
    return await meter.get_trace_transaction(filter_obj)

@method
@async_serialize
async def trace_block(filter_obj):
    return await meter.get_trace_block(filter_obj)



WSURL_NHEADS = 'ws://127.0.0.1:8669/subscriptions/beat'
connections = {}


async def subscribe_headers(parameter):
    hash_object = hashlib.sha224(parameter.encode())
    hex_dig = hash_object.hexdigest()
    if not hex_dig in connections:
        connections[hex_dig] = {'state':1}
    return await handle_socket(hex_dig)
    

@method
@async_serialize
async def eth_subscribe(parameter): 
    if parameter == "newHeads":
        return await subscribe_headers(parameter)
       
       


@method
def eth_unsubscribe(parameter):
    if parameter in connections:
        connections[parameter]['state'] = 0
        


async def handle_socket(key):
        try:
            async with websockets.connect(WSURL_NHEADS) as ws:
                try:
                    
                    if connections[key]['state'] == 1:
                        
                        msg = await ws.recv()
                      
                        return json.loads(msg)
                        
                    elif connections[key]['state'] == 0:
                        del connections[key]
                        return True
                        
                except:
                    await ws.close()
        except:
            return {'error code':404, 'message':'Failed to connect to url'}



from threading import Thread


class ThreadWithReturn(Thread):
    '''
    以多线程方式执行函数, 并且可在线程结束后通过 get 方法获取函数执行结果 (如果函数有返回值)
    '''

    def __init__(self, target=None, args=None, kwargs=None):
        super(ThreadWithReturn, self).__init__(target=target, args=args, kwargs=kwargs)
        self.target = target
        self.args = args
        self.kwargs = kwargs

    def run(self):
        self._return = self.target(*self.args, **self.kwargs)

    def get(self, timeout=None):
        self.join(timeout)
        try:
            return self._return
        except AttributeError:
            raise RuntimeError("Something went wrong.  No `_return` property was set")


def spawn(target, *args, **kwargs):
    thread = ThreadWithReturn(
        target=target,
        args=args,
        kwargs=kwargs,
    )
    thread.daemon = True
    thread.start()
    return thread



from codecs import open
from os import path
from setuptools import (
    setup,
    find_packages,
)

from gear import __version__

here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="meter-gear",
    version=__version__,
    description="An adapter between meter-restful and eth-rpc.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/meterio/meter-gear",
    author="Simon Zhang",
    author_email="zhanghan.simon@gmail.com",
    license="MIT",
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.7',
    ],
    keywords="meter blockchain ethereum",
    packages=find_packages("."),
    include_package_data=True,
    python_requires=">=3.7",
    install_requires=[x.strip() for x in open('requirements.txt')],
    entry_points={
        "console_scripts": [
            "meter-gear=gear.cli:main",
        ],
    }
)



import websockets
import asyncio

async def listen():
    url = "ws://127.0.0.1:8545/ws"

    async with websockets.connect(url) as ws:
        await ws.send("Hello server!")
        while True:
            msg = await ws.recv()
            print(msg)

asyncio.get_event_loop().run_until_complete(listen())